import pyfastnoisesimd
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
import numpy as np


def perlin_occupancygrid(
    w: int, h: int, thresh: float = 0.33, frames: int = None
) -> np.ndarray:
    """Make an occupancyGrid with thresholded Perlin Noise.

    Parameters
    ----------
    w : int
        width of the occupancyGrid
    h : int
        height of the occupancyGrid
    thresh : float, optional
        threshold of the noise. 0.1 produces an occupancyGrid that is mostly free space,
        0.9 produces an occupancyGrid that is mostly obstacle space. by default 0.3
    frames : int, optional
        number of "frames" of occupancygrid. If a number is passed, this will produce a
        3-dimensional occupancygrid, with the first dimension being time. This is
        designed to be used to simulate e.g. a dynamically, but smoothly changing
        occupancyGrid., by default None

    Returns
    -------
    np.ndarray
        occupancyGrid where 1 is occupied and 0 is free space.
    """
    # create noise
    noise = pyfastnoisesimd.Noise()
    noise.noiseType = pyfastnoisesimd.NoiseType(5)
    if frames is not None:
        xynoise = noise.genAsGrid(shape=[frames, w, h])
    else:
        xynoise = noise.genAsGrid(shape=[1, w, h])
        xynoise = np.squeeze(xynoise)
    # normalize to [0,1]
    xynoise -= xynoise.min()
    xynoise = xynoise / (xynoise.max() - xynoise.min())
    # threshold
    xynoise = np.where(xynoise >= thresh, 0, 1)
    return xynoise


if __name__ == "__main__":
    fig = plt.figure()
    ax = fig.add_subplot(111)
    xynoise = perlin_occupancygrid(1000, 1000, frames=1000)
    im = ax.imshow(xynoise[0], cmap="gray_r", interpolation=None)

    def anim_f(n):
        im.set_data(xynoise[n])
        return (im,)

    animation = FuncAnimation(
        fig, anim_f, frames=xynoise.shape[0], interval=10, blit=True
    )
    plt.show()



import numpy as np
from matplotlib.collections import LineCollection
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.patches import Circle
from tqdm import tqdm
from scipy.ndimage import binary_dilation
from rrt import random_point_og, RRT


DEFAULT_APPEARANCE = {
    "tree_color": "slategrey",
    "tree_alpha": 0.2,
    "og_cmap": "gist_gray_r",
    "goal_marker": "*",
    "goal_color": "seagreen",
    "goal_circ_color": "seagreen",
    "start_marker": "o",
    "start_color": "red",
    "path_color": "blue",
}


class DynamicEnvironmentAnimation(object):
    def __init__(
        self,
        movespeed: float,
        r_within_goal: float,
        rrtobj: RRT,
        appearance: dict = DEFAULT_APPEARANCE,
        buffer_size: int = None,
    ):
        self.movespeed = movespeed
        # radius within which we have "found" the goal
        self.r_within_goal = r_within_goal
        # RRT object. can be RRT star, informed, etc. Must have `make` method
        # which can take a start and goal and return a tree and a goal vector
        self.rrtobj = rrtobj
        # size of the "buffer" around obstacles. this is to avoid getting stuck
        # in the obstacles. if the environment changes quickly, or if velocity is
        # large, this should be a larger value. But setting a very large value
        # will prevent optimal paths from being found.
        if buffer_size is not None:
            self.buffer_size = buffer_size
        else:
            self.buffer_size = int(movespeed / 2)
        self.appearance = appearance

    @staticmethod
    def clamp(xy, shape):
        x = max(0, min(xy[0], shape[0] - 1))
        y = max(0, min(xy[1], shape[1] - 1))
        return np.array([x, y])

    def simulate_dynamic_goals(
        self,
        og_3d,
    ):
        frames = og_3d.shape[0]
        xstart = random_point_og(og_3d[0])
        xgoal = random_point_og(og_3d[0])
        current_position = xstart

        goals, positions = np.empty((frames, 2)), np.empty((frames, 2))
        paths, trees = [], []

        print("Simulating Dynamic Mission...")
        for fi in tqdm(range(frames), desc="frames"):
            og = og_3d[fi]
            # if we are at the goal, get a new goal
            if np.linalg.norm(current_position - xgoal) < self.r_within_goal:
                xgoal = random_point_og(og)
            current_position = self.clamp(current_position, og.shape)

            # ok so we kinda want to avoid getting stuck in the obstacles so
            # we'll do a binary dilation to get a bit of a buffer around objects
            # before computing the plan, but we remove only the buffer near us.
            bsize = int(self.movespeed / 2)
            og_dilated = binary_dilation(og, iterations=bsize)
            buffer_reg = og_dilated - og
            for i in range(bsize * 2):
                for j in range(bsize * 2):
                    hole = self.clamp(current_position + np.array([i, j]), og.shape)
                    # set as free
                    buffer_reg[hole[0], hole[1]] = 0
            og = og | buffer_reg
            if og[current_position[0], current_position[1]] == 1:
                in_obs = True
            else:
                in_obs = False
            self.rrtobj.set_og(og)
            T, gv = self.rrtobj.make(current_position, xgoal)
            path = self.rrtobj.path_points(T, self.rrtobj.route2gv(T, gv))

            # populate lists
            goals[fi] = xgoal
            positions[fi] = current_position
            paths.append(path)
            tree = []
            for e1, e2 in T.edges:
                p1, p2 = T.nodes[e1]["pt"], T.nodes[e2]["pt"]
                tree.append([p1, p2])
            trees.append(np.array(tree))

            if path.shape[0] != 0 and not in_obs:
                # follow current plan
                v = path[0, 1, :] - path[0, 0, :]
                angle = np.arctan2(v[1], v[0])
                # update simulation
                current_position += np.int64(
                    np.array([np.cos(angle), np.sin(angle)]) * self.movespeed
                )

        return goals, positions, paths, trees

    def animate(self, goals, positions, paths, trees, og_3d):
        fig, ax = plt.subplots()
        ax.set_xlim(0, og_3d.shape[1])
        ax.set_ylim(0, og_3d.shape[2])
        ax.set_aspect("equal")

        pathlc = LineCollection(
            [], color=self.appearance["path_color"], zorder=3, linewidth=2
        )
        treelc = LineCollection(
            [],
            color=self.appearance["tree_color"],
            zorder=2,
            alpha=self.appearance["tree_alpha"],
        )
        im = ax.imshow(
            og_3d[0].T,
            cmap=self.appearance["og_cmap"],
            vmin=0,
            vmax=1,
            interpolation=None,
        )

        goal_sc = ax.scatter(
            [],
            [],
            marker=self.appearance["goal_marker"],
            color=self.appearance["goal_color"],
            s=40,
            zorder=4,
        )
        goalcircle = Circle(
            [0, 0],
            radius=self.r_within_goal,
            fill=None,
            ec=self.appearance["goal_circ_color"],
            zorder=3,
        )
        start_sc = ax.scatter(
            [],
            [],
            marker=self.appearance["start_marker"],
            color=self.appearance["start_color"],
            s=40,
            zorder=4,
        )

        def init():
            ax.add_collection(pathlc)
            ax.add_collection(treelc)
            ax.add_artist(goalcircle)
            return im, pathlc, treelc, goal_sc, goalcircle, start_sc

        def update(i):
            pathlc.set_segments(paths[i])
            treelc.set_segments(trees[i])
            im.set_data(og_3d[i].T)
            goal_sc.set_offsets([goals[i, 0], goals[i, 1]])
            goalcircle.center = goals[i]
            start_sc.set_offsets([positions[i, 0], positions[i, 1]])
            return im, pathlc, treelc, goal_sc, goalcircle, start_sc

        anim = FuncAnimation(
            fig,
            update,
            init_func=init,
            frames=goals.shape[0],
            interval=50,
            blit=True,
        )
        return anim

    def make_animation(self, og_3d):
        results = self.simulate_dynamic_goals(og_3d)
        return self.animate(*results, og_3d)


if __name__ == "__main__":
    import rrt
    from oggen import perlin_occupancygrid

    w, h = 100, 100
    frames = 200
    og_3d = perlin_occupancygrid(w, h, 0.33, frames)
    rrts = rrt.RRTStarInformed(og_3d[0], n=1000, r_rewire=25, r_goal=12.5, pbar=False)

    dyn_env_anim = DynamicEnvironmentAnimation(7.0, 5.0, rrts)
    anim = dyn_env_anim.make_animation(og_3d)
    # anim.save("animation.mp4", fps=30, dpi=300)
    plt.show()



import numpy as np
import numba as nb
from math import sqrt
from tqdm import tqdm
import networkx as nx
from typing import Tuple, List
from collections import defaultdict


@nb.njit(fastmath=True)
def r2norm(x):
    return sqrt(x[0] * x[0] + x[1] * x[1])


def random_point_og(og: np.ndarray) -> np.ndarray:
    """Get a random point in free space from the occupancyGrid.

    Parameters
    ----------
    og : np.ndarray
        occupancyGrid. 1 is obstacle, 0 is free space.

    Returns
    -------
    np.ndarray
        randomly-sampled point
    """
    free = np.argwhere(og == 0)
    return free[np.random.randint(0, free.shape[0])]


############# RRT BASE CLASS ##################################################


class RRT(object):
    """base class containing common RRT methods"""

    def __init__(
        self,
        og: np.ndarray,
        n: int,
        costfn: callable = None,
        every: int = 10,
        pbar: bool = True,
    ):
        # whether to display a progress bar
        self.pbar = pbar
        # every n tries, attempt to go to goal
        self.every = every
        # array containing vertex points
        self.n = n
        # occupancyGrid free space
        self.free = np.argwhere(og == 0)
        self.og = og

        # define simple r2norm cost function if
        # no cost function is passed in

        if costfn is None:

            def costfn(
                vcosts: np.ndarray,
                points: np.ndarray,
                v: int,
                x: np.ndarray,
            ) -> float:
                return vcosts[v] + r2norm(points[v] - x)

        self.cost = costfn
        self.not_a_point = [np.inf, np.inf]
        self.not_a_dist = np.inf

    def route2gv(self, T: nx.DiGraph, gv) -> List[int]:
        return nx.shortest_path(T, source=0, target=gv, weight="dist")

    def path_points(self, T: nx.DiGraph, path: list) -> np.ndarray:
        lines = []
        for i in range(len(path) - 1):
            lines.append([T.nodes[path[i]]["pt"], T.nodes[path[i + 1]]["pt"]])
        return np.array(lines)

    @staticmethod
    def near(points: np.ndarray, x: np.ndarray) -> np.ndarray:
        # vector from x to all points
        p2x = points - x
        # norm of that vector
        dist = np.linalg.norm(p2x, axis=1)
        # sorted norms
        sorted = np.argsort(dist)
        return sorted

    @staticmethod
    def within(points: np.ndarray, x: np.ndarray, r: float) -> np.ndarray:
        # vector from x to all points
        p2x = points - x
        # dot dist with self to get r2
        d2 = p2x[:, 0] * p2x[:, 0] + p2x[:, 1] * p2x[:, 1]
        # get indices of points within r2
        near_idx = np.argwhere(d2 < r * r)
        return np.atleast_1d(np.squeeze(near_idx))

    @staticmethod
    @nb.njit()
    def collisionfree(og, a, b) -> bool:
        """calculate linear collision on occupancyGrid between points a, b"""
        x0 = a[0]
        y0 = a[1]
        x1 = b[0]
        y1 = b[1]
        dx = abs(x1 - x0)
        if x0 < x1:
            sx = 1
        else:
            sx = -1
        dy = -abs(y1 - y0)
        if y0 < y1:
            sy = 1
        else:
            sy = -1
        err = dx + dy
        while True:
            if og[x0, y0] != 0:
                return False
            elif x0 == x1 and y0 == y1:
                return True
            else:
                e2 = 2 * err
                if e2 >= dy:
                    err += dy
                    x0 += sx
                if e2 <= dx:
                    err += dx
                    y0 += sy

    def sample_all_free(self):
        """sample uniformly from free space in the occupancyGrid."""
        return self.free[np.random.choice(self.free.shape[0])]

    def make(self, xstart: np.ndarray, xgoal: np.ndarray):
        raise NotImplementedError

    def set_og(self, og_new: np.ndarray):
        """update occupancyGrid"""
        self.og = og_new
        self.free = np.argwhere(og_new == 0)

    def set_n(self, n: int):
        """update n"""
        self.n = n

    def set_every(self, every: int):
        """update every"""
        self.every = every

    def go2goal(self, vcosts, points, xgoal, j, children, parents):
        # cost for all existing points
        costs = np.empty(vcosts.shape)
        for i in range(points.shape[0]):
            costs[i] = self.cost(vcosts, points, i, xgoal)

        found_goal = False
        for idx in np.argsort(costs):
            if self.collisionfree(self.og, points[idx], xgoal):
                vgoal = j
                points = np.concatenate((points, xgoal[np.newaxis, :]), axis=0)
                vcosts = np.concatenate((vcosts, [costs[idx]]), axis=0)
                points[vgoal] = xgoal
                vcosts[vgoal] = costs[idx]
                children[idx].append(vgoal)
                parents[vgoal] = idx
                found_goal = True
                break
        if not found_goal:
            # shrink points array
            dists = np.linalg.norm(points - xgoal)
            vgoal = np.argmin(dists)
        return vgoal, children, parents, points, vcosts

    def build_graph(self, vgoal, points, parents, vcosts):
        assert points.max() < self.not_a_point[0] - 1
        # build graph
        T = nx.DiGraph()
        T.add_node(vgoal, pt=points[vgoal])
        for i, p in enumerate(points):
            T.add_node(i, pt=p)

        for child, parent in parents.items():
            if parent is not None:
                p1, p2 = points[parent], points[child]
                dist = r2norm(p2 - p1)
                T.add_edge(parent, child, dist=dist, cost=vcosts[child])
        return T


############# RRT STANDARD ####################################################


class RRTStandard(RRT):
    def __init__(
        self,
        og: np.ndarray,
        n: int,
        costfn: callable = None,
        every=100,
        pbar=True,
    ):
        super().__init__(og, n, costfn=costfn, every=every, pbar=pbar)

    def make(self, xstart: np.ndarray, xgoal: np.ndarray) -> Tuple[nx.DiGraph, int]:
        sampled = set()
        points = np.full((self.n, 2), dtype=int, fill_value=self.not_a_point)
        vcosts = np.full((self.n,), fill_value=self.not_a_dist)
        parents, children = {}, defaultdict(list)
        points[0] = xstart
        vcosts[0] = 0
        parents[0] = None
        i, j = 0, 1
        if self.pbar:
            pbar = tqdm(total=self.n)

        while i < self.n:
            if self.pbar:
                pbar.update(1)
            xnew = self.sample_all_free()
            vnearest = self.near(points, xnew)[0]
            xnearest = points[vnearest]
            nocoll = self.collisionfree(self.og, xnearest, xnew)
            if nocoll and tuple(xnew) not in sampled and j != self.n:
                sampled.add(tuple(xnew))
                # check least cost path to xnew
                vbest = vnearest
                # store new point
                vnew = j
                points[vnew] = xnew
                vcosts[vnew] = self.cost(vcosts, points, vbest, xnew)
                # store new edge
                children[vbest].append(vnew)
                parents[vnew] = vbest
                j += 1
            i += 1

        # go to goal if possible
        vgoal, children, parents, points, vcosts = self.go2goal(
            vcosts, points, xgoal, j, children, parents
        )

        # build graph
        T = self.build_graph(vgoal, points, parents, vcosts)

        return T, vgoal


########### RRT STAR ##########################################################


class RRTStar(RRT):
    def __init__(
        self,
        og: np.ndarray,
        n: int,
        r_rewire: float,
        costfn: callable = None,
        every=100,
        pbar=True,
    ):
        super().__init__(og, n, costfn=costfn, every=every, pbar=pbar)
        self.r_rewire = r_rewire

    def make(self, xstart: np.ndarray, xgoal: np.ndarray):
        sampled = set()
        points = np.full((self.n, 2), dtype=int, fill_value=self.not_a_point)
        vcosts = np.full((self.n,), fill_value=self.not_a_dist)
        children, parents = defaultdict(list), {}
        points[0] = xstart
        vcosts[0] = 0
        parents[0] = None
        i, j = 0, 1
        if self.pbar:
            pbar = tqdm(total=self.n)

        while i < self.n:
            if self.pbar:
                pbar.update(1)

            xnew = self.sample_all_free()
            vnearest = self.near(points, xnew)[0]
            xnearest = points[vnearest]

            nocoll = self.collisionfree(self.og, xnearest, xnew)
            if nocoll and tuple(xnew) not in sampled and j != self.n:
                sampled.add(tuple(xnew))

                # check least cost path to xnew
                vbest = vnearest
                cbest = self.cost(vcosts, points, vbest, xnew)
                vnear = self.within(points, xnew, self.r_rewire)

                for vn in vnear:
                    xn = points[vn]
                    cn = self.cost(vcosts, points, vn, xnew)
                    if cn < cbest:
                        if self.collisionfree(self.og, xn, xnew):
                            vbest = vn
                            cbest = cn

                # store new point
                vnew = j
                points[vnew] = xnew
                vcosts[vnew] = cbest
                # store new edge
                parents[vnew] = vbest
                children[vbest].append(vnew)

                # tree rewire
                for vn in vnear:
                    xn = points[vn]
                    cn = vcosts[vn]
                    cmaybe = self.cost(vcosts, points, vn, xnew)
                    if cmaybe < cn:
                        if self.collisionfree(self.og, xn, xnew):
                            parent = parents[vn]
                            if parent is not None:
                                # reassign parent
                                try:
                                    children[parent].remove(vn)
                                    parents[vn] = vnew
                                    vcosts[vn] = cmaybe
                                except ValueError:
                                    pass
                j += 1
            i += 1
        # go to goal if possible
        vgoal, children, parents, points, vcosts = self.go2goal(
            vcosts, points, xgoal, j, children, parents
        )
        # build graph
        T = self.build_graph(vgoal, points, parents, vcosts)

        return T, vgoal


########### RRT STAR INFORMED #################################################


class RRTStarInformed(RRT):
    def __init__(
        self,
        og: np.ndarray,
        n: int,
        r_rewire: float,
        r_goal: float,
        costfn: callable = None,
        every: int = 100,
        pbar: bool = True,
    ):
        super().__init__(og, n, costfn=costfn, every=every, pbar=pbar)
        self.r_rewire = r_rewire
        self.r_goal = r_goal
        # store the ellipses for plotting later
        self.ellipses = {}

    @staticmethod
    def unitball():
        """draw a point from a uniform distribution bounded by the ball:
        U(x1, x2) ~ 1 > (x1)^2 + (x2)^2"""
        r = np.random.uniform(0, 1)
        theta = 2 * np.pi * np.random.uniform(0, 1)
        x = np.sqrt(r) * np.cos(theta)
        y = np.sqrt(r) * np.sin(theta)
        unif = np.array([x, y])
        return unif

    def sample_ellipse(self, xstart, xgoal, c, clamp=True):
        xcent = (xstart + xgoal) / 2
        CL = self.get_ellipse_xform(xstart, xgoal, c)
        xball = self.unitball()
        x, y = tuple(np.dot(CL, xball) + xcent)

        if clamp:
            # clamp to finite og
            x = int(max(0, min(self.og.shape[0] - 1, x)))
            y = int(max(0, min(self.og.shape[1] - 1, y)))
        return np.array((x, y))

    def rotation_to_world_frame(self, xstart, xgoal):
        """calculate the rotation matrix from the world-frame to the frame given
        by the hyperellipsoid with focal points at xf1=xstart and xf2=xgoal. a unit
        ball multiplied by this matrix will produce an oriented ellipsoid with those
        focal points."""
        a1 = np.atleast_2d((xgoal - xstart) / np.linalg.norm(xgoal - xstart))
        M = np.outer(a1, np.atleast_2d([1, 0]))
        try:
            U, _, V = np.linalg.svd(M)
        except np.linalg.LinAlgError:
            # handle SVD not converging
            U, _, V = np.linalg.svd(M, full_matrices=False)
        return U @ np.diag([np.linalg.det(U), np.linalg.det(V)]) @ V.T

    def get_ellipse_xform(self, xstart, xgoal, cmax):
        """transform vector in unit plane to ellipse plane"""
        # rotation matrix
        C = self.rotation_to_world_frame(xstart, xgoal)
        # scale by major axis r1, minor axis r2.
        r1 = cmax / 2
        d2 = np.dot((xstart - xgoal).T, (xstart - xgoal))
        r2 = np.sqrt(abs(cmax * cmax - d2)) / 2
        L = np.diag([r1, r2])
        # dot rotation with scale to get final matrix
        return np.dot(C, L)

    @staticmethod
    def least_cost(vcosts, vsoln):
        if len(vsoln) == 1:
            return vsoln[0], vcosts[vsoln[0]]
        else:
            idx_cmin = np.argmin(vcosts[vsoln])
            return vsoln[idx_cmin], vcosts[vsoln[idx_cmin]]

    @staticmethod
    def rad2deg(a):
        return a * 180 / np.pi

    def get_ellipse_for_plt(
        self, xstart, xgoal, cmax
    ) -> Tuple[np.ndarray, float, float, float]:
        xcent = (xgoal + xstart) / 2
        # get the rotation
        CL = self.get_ellipse_xform(xstart, xgoal, cmax)
        # apply the rotation to find major axis, minor axis, angle
        a = np.dot(CL, np.array([1, 0]))
        b = np.dot(CL, np.array([0, 1]))
        majax = 2 * np.linalg.norm(a)
        minax = 2 * np.linalg.norm(b)
        ang = self.rad2deg(np.arctan2((a)[1], (a)[0]))
        return xcent, majax, minax, ang

    def make(self, xstart: np.ndarray, xgoal: np.ndarray):

        vsoln = []
        sampled = set()
        points = np.full((self.n, 2), dtype=int, fill_value=self.not_a_point)
        vcosts = np.full((self.n,), fill_value=self.not_a_dist)
        parents, children = {}, defaultdict(list)
        points[0] = xstart
        vcosts[0] = 0
        parents[0] = None
        i, j = 0, 1
        if self.pbar:
            pbar = tqdm(total=self.n)

        while i < self.n:
            if self.pbar:
                pbar.update(1)

            # either sample from free or sample from ellipse depending on vsoln
            if len(vsoln) == 0:
                xnew = self.sample_all_free()
            else:
                vbest, cbest = self.least_cost(vcosts, list(vsoln))
                cbest += r2norm(xgoal - points[vbest])
                xnew = self.sample_ellipse(xstart, xgoal, cbest)
                self.ellipses[j] = self.get_ellipse_for_plt(xstart, xgoal, cbest)

            vnearest = self.near(points, xnew)[0]
            xnearest = points[vnearest]

            nocoll = self.collisionfree(self.og, xnearest, xnew)
            if nocoll and tuple(xnew) not in sampled and j != self.n:
                sampled.add(tuple(xnew))

                # check least cost path to xnew
                vbest = vnearest
                cbest = self.cost(vcosts, points, vbest, xnew)
                vnear = self.within(points, xnew, self.r_rewire)

                for vn in vnear:
                    xn = points[vn]
                    cn = self.cost(vcosts, points, vn, xnew)
                    if cn < cbest:
                        if self.collisionfree(self.og, xn, xnew):
                            vbest = vn
                            cbest = cn

                # store new point
                vnew = j
                points[vnew] = xnew
                vcosts[vnew] = cbest
                # store new edge
                parents[vnew] = vbest
                children[vbest].append(vnew)

                # tree rewire
                for vn in vnear:
                    xn = points[vn]
                    cn = vcosts[vn]
                    cmaybe = self.cost(vcosts, points, vn, xnew)
                    if cmaybe < cn:
                        if self.collisionfree(self.og, xn, xnew):
                            parent = parents[vn]
                            if parent is not None:
                                children[parent].remove(vn)
                                parents[vn] = vnew
                                vcosts[vn] = cmaybe
                # check goal
                if r2norm(xnew - xgoal) < self.r_goal:
                    vsoln.append(vnew)

                j += 1
            i += 1

        # go to goal if possible
        vgoal, children, parents, points, vcosts = self.go2goal(
            vcosts, points, xgoal, j, children, parents
        )

        # build graph
        T = self.build_graph(vgoal, points, parents, vcosts)

        return T, vgoal


if __name__ == "__main__":
    from oggen import perlin_occupancygrid
    import matplotlib.pyplot as plt
    from plots import plot_og, plot_rrt_lines, plot_path, plot_start_goal

    # create occupancy Grid
    w, h = 800, 450
    og = perlin_occupancygrid(w, h)

    # create planner
    rrtplanner = RRTStarInformed(og, n=2000, r_rewire=64, r_goal=12)
    xstart = random_point_og(og)
    xgoal = random_point_og(og)

    T, gv = rrtplanner.make(xstart, xgoal)
    pathlines = rrtplanner.route2gv(T, gv)
    pathlines = rrtplanner.path_points(T, pathlines)

    fig = plt.figure(figsize=(16, 9))
    ax = fig.add_subplot(111)
    plot_og(ax, og)
    plot_start_goal(ax, xstart, xgoal)
    plot_rrt_lines(ax, T, color_costs=True)
    plot_path(ax, pathlines)
    plt.show()



from matplotlib.colors import Normalize
from matplotlib.collections import LineCollection
from matplotlib import cm
from matplotlib.patches import Ellipse
import numpy as np
from matplotlib.axes import Axes


def remove_axticks(ax):
    ax.set_xticks([])
    ax.set_yticks([])


def plot_og(ax: Axes, og: np.ndarray, cmap: str = "Greys", vmin=0, vmax=1):
    norm = Normalize(vmin=vmin, vmax=vmax)
    ax.imshow(og.T, cmap=cmap, norm=norm, origin="lower", interpolation=None)


def plot_start_goal(ax, xstart, xgoal):
    ax.scatter(xstart[0], xstart[1], c="r", s=50, label="start", marker="o")
    ax.scatter(xgoal[0], xgoal[1], c="b", s=50, label="goal", marker="*")


def plot_rrt_lines(ax, T, color_costs=True, cmap="viridis", color="tan", alpha=1.0):
    lines = []
    costs = []
    for e1, e2 in T.edges():
        lines.append((T.nodes[e1]["pt"], T.nodes[e2]["pt"]))
        costs.append(T.edges[e1, e2]["cost"])
    if color_costs:
        norm = Normalize(vmin=min(costs), vmax=max(costs))
        colors = cm.get_cmap(cmap)(norm(costs))
        lc = LineCollection(lines, colors=colors, alpha=alpha)
    else:
        lc = LineCollection(lines, color=color, alpha=alpha)
    ax.add_collection(lc)


def plot_rrt_points(ax, T):
    for n in T.nodes:
        ax.scatter(T.nodes[n]["pt"][0], T.nodes[n]["pt"][1], c="k", s=10, marker="o")


def plot_ellipses(ax, ellipses, cmap="RdPu"):
    if len(ellipses) == 0:
        return None
    colors = [j for j in ellipses.keys()]
    norm = Normalize(vmin=min(colors), vmax=max(colors))
    colors = cm.get_cmap(cmap)(norm(colors))
    for e, c in zip(ellipses.values(), colors):
        ax.add_artist(Ellipse(*e, fill=None, color=c, zorder=10))


def plot_path(ax, pathlines: np.ndarray, zorder=10):
    lc = LineCollection(pathlines, color="r", linewidth=2, zorder=zorder)
    ax.add_collection(lc)


if __name__ == "__main__":
    from matplotlib import pyplot as plt
    import rrt, oggen

    axs = []
    dpi, figsize = 120, (12, 6)

    # create world
    w, h = 700, 500
    n = 2500
    r_rewire, r_goal = 64, 32
    og = oggen.perlin_occupancygrid(w, h)
    xstart = rrt.random_point_og(og)
    xgoal = rrt.random_point_og(og)

    fig, axs = plt.subplots(2, 2, figsize=figsize, dpi=dpi, tight_layout=True)

    axs[0, 0].set_title("Empty OccupancyGrid")

    rrt2 = rrt.RRTStandard(og, n)
    T, gv = rrt2.make(xstart, xgoal)
    path = rrt2.path_points(T, rrt2.route2gv(T, gv))
    plot_rrt_lines(axs[0, 1], T)
    plot_path(axs[0, 1], path)
    axs[0, 1].set_title("Standard RRT")

    rrt3 = rrt.RRTStar(og, n, r_rewire)
    T, gv = rrt3.make(xstart, xgoal)
    plot_rrt_lines(axs[1, 0], T)
    path = rrt3.path_points(T, rrt3.route2gv(T, gv))
    plot_path(axs[1, 0], path)
    axs[1, 0].set_title("RRT*")

    rrt4 = rrt.RRTStarInformed(og, n, r_rewire, r_goal)
    T, gv = rrt4.make(xstart, xgoal)
    path = rrt4.path_points(T, rrt4.route2gv(T, gv))
    plot_rrt_lines(axs[1, 1], T)
    plot_ellipses(axs[1, 1], rrt4.ellipses)
    plot_path(axs[1, 1], path)
    axs[1, 1].set_title("RRT* Informed")

    for ax in axs.flatten():
        remove_axticks(ax)
        plot_og(ax, og)
        plot_start_goal(ax, xstart, xgoal)

    # fig.savefig("./figure.png", dpi=200)

    plt.show()
